@@ -3,7 +3,7 @@ from click_default_group import DefaultGroup
 import json
 import os
 import pathlib
-from playwright.sync_api import sync_playwright, Error
+from playwright.sync_api import sync_playwright, Error, TimeoutError
 from runpy import run_module
 import secrets
 import sys
@@ -13,6 +13,19 @@ import yaml
 
 from shot_scraper.utils import filename_for_url, url_or_file_path
 
+BROWSERS = ("chromium", "firefox", "chrome", "chrome-beta")
+
+
+def browser_option(fn):
+    click.option(
+        "--browser",
+        "-b",
+        default="chromium",
+        type=click.Choice(BROWSERS, case_sensitive=False),
+        help="Which browser to use",
+    )(fn)
+    return fn
+
 
 @click.group(
     cls=DefaultGroup,
@@ -72,6 +85,11 @@ def cli():
 @click.option(
     "--wait", type=int, help="Wait this many milliseconds before taking the screenshot"
 )
+@click.option(
+    "--timeout",
+    type=int,
+    help="Wait this many milliseconds before failing",
+)
 @click.option(
     "-i",
     "--interactive",
@@ -83,6 +101,7 @@ def cli():
     is_flag=True,
     help="Interact mode with developer tools",
 )
+@browser_option
 def shot(
     url,
     auth,
@@ -95,8 +114,10 @@ def shot(
     retina,
     quality,
     wait,
+    timeout,
     interactive,
     devtools,
+    browser,
 ):
     """
     Take a single screenshot of a page or portion of a page.
@@ -134,14 +155,21 @@ def shot(
         "height": height,
         "quality": quality,
         "wait": wait,
+        "timeout": timeout,
         "padding": padding,
         "retina": retina,
     }
     interactive = interactive or devtools
     with sync_playwright() as p:
         use_existing_page = False
-        context, browser = _browser_context(
-            p, auth, interactive=interactive, devtools=devtools, retina=retina
+        context, browser_obj = _browser_context(
+            p,
+            auth,
+            interactive=interactive,
+            devtools=devtools,
+            retina=retina,
+            browser=browser,
+            timeout=timeout,
         )
         if interactive or devtools:
             use_existing_page = True
@@ -152,26 +180,49 @@ def shot(
                 "Hit <enter> to take the shot and close the browser window:", err=True
             )
             input()
-        if output == "-":
-            shot = take_shot(
-                context, shot, return_bytes=True, use_existing_page=use_existing_page
-            )
-            sys.stdout.buffer.write(shot)
-        else:
-            shot["output"] = str(output)
-            shot = take_shot(context, shot, use_existing_page=use_existing_page)
-        browser.close()
+        try:
+            if output == "-":
+                shot = take_shot(
+                    context,
+                    shot,
+                    return_bytes=True,
+                    use_existing_page=use_existing_page,
+                )
+                sys.stdout.buffer.write(shot)
+            else:
+                shot["output"] = str(output)
+                shot = take_shot(context, shot, use_existing_page=use_existing_page)
+        except TimeoutError as e:
+            raise click.ClickException(str(e))
+        browser_obj.close()
 
 
-def _browser_context(p, auth, interactive=False, devtools=False, retina=False):
-    browser = p.chromium.launch(headless=not interactive, devtools=devtools)
+def _browser_context(
+    p,
+    auth,
+    interactive=False,
+    devtools=False,
+    retina=False,
+    browser="chromium",
+    timeout=None,
+):
+    browser_kwargs = dict(headless=not interactive, devtools=devtools)
+    if browser == "chromium":
+        browser_obj = p.chromium.launch(**browser_kwargs)
+    elif browser == "firefox":
+        browser_obj = p.firefox.launch(**browser_kwargs)
+    else:
+        browser_kwargs["channel"] = browser
+        browser_obj = p.chromium.launch(**browser_kwargs)
     context_args = {}
     if auth:
         context_args["storage_state"] = json.load(auth)
     if retina:
         context_args["device_scale_factor"] = 2
-    context = browser.new_context(**context_args)
-    return context, browser
+    context = browser_obj.new_context(**context_args)
+    if timeout:
+        context.set_default_timeout(timeout)
+    return context, browser_obj
 
 
 @cli.command()
@@ -183,7 +234,14 @@ def _browser_context(p, auth, interactive=False, devtools=False, retina=False):
     help="Path to JSON authentication context file",
 )
 @click.option("--retina", is_flag=True, help="Use device scale factor of 2")
-def multi(config, auth, retina):
+@click.option(
+    "--timeout",
+    type=int,
+    help="Wait this many milliseconds before failing",
+)
+@click.option("--fail-on-error", is_flag=True, help="Fail noisily on error")
+@browser_option
+def multi(config, auth, retina, timeout, fail_on_error, browser):
     """
     Take multiple screenshots, defined by a YAML file
 
@@ -203,10 +261,19 @@ def multi(config, auth, retina):
     if not isinstance(shots, list):
         raise click.ClickException("YAML file must contain a list")
     with sync_playwright() as p:
-        context, browser = _browser_context(p, auth, retina=retina)
+        context, browser_obj = _browser_context(
+            p, auth, retina=retina, browser=browser, timeout=timeout
+        )
         for shot in shots:
-            take_shot(context, shot)
-        browser.close()
+            try:
+                take_shot(context, shot)
+            except TimeoutError as e:
+                if fail_on_error:
+                    raise click.ClickException(str(e))
+                else:
+                    click.echo(str(e), err=True)
+                    continue
+        browser_obj.close()
 
 
 @cli.command()
@@ -234,13 +301,13 @@ def accessibility(url, auth, output, javascript):
     """
     url = url_or_file_path(url, _check_and_absolutize)
     with sync_playwright() as p:
-        context, browser = _browser_context(p, auth)
+        context, browser_obj = _browser_context(p, auth)
         page = context.new_page()
         page.goto(url)
         if javascript:
             _evaluate_js(page, javascript)
         snapshot = page.accessibility.snapshot()
-        browser.close()
+        browser_obj.close()
     output.write(json.dumps(snapshot, indent=4))
     output.write("\n")
 
@@ -268,7 +335,8 @@ def accessibility(url, auth, output, javascript):
     default="-",
     help="Save output JSON to this file",
 )
-def javascript(url, javascript, input, auth, output):
+@browser_option
+def javascript(url, javascript, input, auth, output, browser):
     """
     Execute JavaScript against the page and return the result as JSON
 
@@ -298,11 +366,11 @@ def javascript(url, javascript, input, auth, output):
         javascript = input.read()
     url = url_or_file_path(url, _check_and_absolutize)
     with sync_playwright() as p:
-        context, browser = _browser_context(p, auth)
+        context, browser_obj = _browser_context(p, auth, browser=browser)
         page = context.new_page()
         page.goto(url)
         result = _evaluate_js(page, javascript)
-        browser.close()
+        browser_obj.close()
     output.write(json.dumps(result, indent=4, default=str))
     output.write("\n")
 
@@ -344,7 +412,7 @@ def pdf(url, auth, output, javascript, wait, media_screen, landscape):
     if output is None:
         output = filename_for_url(url, ext="pdf", file_exists=os.path.exists)
     with sync_playwright() as p:
-        context, browser = _browser_context(p, auth)
+        context, browser_obj = _browser_context(p, auth)
         page = context.new_page()
         page.goto(url)
         if wait:
@@ -370,19 +438,30 @@ def pdf(url, auth, output, javascript, wait, media_screen, landscape):
                 "Screenshot of '{}' written to '{}'".format(url, output), err=True
             )
 
-        browser.close()
+        browser_obj.close()
 
 
 @cli.command()
-def install():
+@click.option(
+    "--browser",
+    "-b",
+    default="chromium",
+    type=click.Choice(BROWSERS, case_sensitive=False),
+    help="Which browser to install",
+)
+def install(browser):
     """
-    Install Playwright browser needed by this tool.
+    Install the Playwright browser needed by this tool.
 
     Usage:
 
         shot-scraper install
+
+    Or for browsers other than the Chromium default:
+
+        shot-scraper install -b firefox
     """
-    sys.argv = ["playwright", "install", "chromium"]
+    sys.argv = ["playwright", "install", browser]
     run_module("playwright", run_name="__main__")
 
 
@@ -430,7 +509,12 @@ def _check_and_absolutize(filepath):
     return False
 
 
-def take_shot(context_or_page, shot, return_bytes=False, use_existing_page=False):
+def take_shot(
+    context_or_page,
+    shot,
+    return_bytes=False,
+    use_existing_page=False,
+):
     url = shot.get("url") or ""
     if not url:
         raise ShotError("url is required")
@@ -465,12 +549,10 @@ def take_shot(context_or_page, shot, return_bytes=False, use_existing_page=False
         if shot.get("height"):
             full_page = False
 
-    if not use_existing_page:
-        page.goto(url)
+    page.goto(url)
 
     if wait:
         time.sleep(wait / 1000)
-    message = ""
     javascript = shot.get("javascript")
     if javascript:
         _evaluate_js(page, javascript)
